unit WGenForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, ExtCtrls,
  MMSystem, Menus;

const
  MaxBufSize = 11050;
type
  PStereoBuf = ^TStereoBuf;
  TStereoBuf =array[0..2 * MaxBufSize - 1] of SmallInt;

  TDualWaveGen = class(TForm)
    Panel1: TPanel;
    RFreqLabel: TLabel;
    LFreqBar: TTrackBar;
    LFreqLabel: TLabel;
    Label21: TLabel;
    RFreqBar: TTrackBar;
    CloseBtn: TButton;
    PlayBtn: TButton;
    Panel2: TPanel;
    PaintBox2: TPaintBox;
    PaintBox3: TPaintBox;
    HrznScrollBar: TScrollBar;
    PaintBox1: TPaintBox;
    LeftVolBar: TTrackBar;
    RightVolBar: TTrackBar;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    procedure PaintBox1Paint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PlayBtnClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure CloseBtnClick(Sender: TObject);
    procedure LFreqBarChange(Sender: TObject);
    procedure RFreqBarChange(Sender: TObject);
    procedure HrznScrollBarChange(Sender: TObject);
    procedure LeftVolBarChange(Sender: TObject);
    procedure RightVolBarChange(Sender: TObject);
  private
    { Private 錾 }
    FWaveOut: HWaveOut;           // handle to wave out device
    FWaveFmt: TWaveFormatEx;      // wave format descriptor
    Playing: boolean;
    FWaveHdr1: PWaveHdr;         // 1st wave heder
    FWaveHdr2: PWaveHdr;         // 2nd wave heder
    FWaveBuf1: PStereoBuf;       // 1st wave data; stereo 16bits
    FWaveBuf2: PStereoBuf;       // 2nd wave data; stereo 16bits
    FreeBuf: PStereoBuf;

    SinTab: array of double;
    SinTabCount: Integer;
    LVolume: double;              // range [0, 1]
    RVolume: double;              // range [0, 1]
    SamplingRate: Integer;
    LeftFreq: Integer;
    RightFreq: Integer;
    MinFreq : integer;
    LAngle: Integer;
    RAngle: Integer;
    LeftPoints: array of TPoint;
    RightPoints: array of TPoint;
    BeatPoints: array of TPoint;
    ShutOff: Boolean;
    Closed: Boolean;
    SoundChanged: boolean;
    SampleBits: Integer;
    procedure MMWomOpen(var Msg: TMessage);  message mm_wom_open;
    procedure MMWomDone(var Msg: TMessage);  message mm_wom_done;
    procedure MMWomClose(var Msg: TMessage);  message mm_wom_close;
    procedure SynthWave;
    procedure ToneChange;
    procedure DrawWave;
    procedure EraseWave;
    procedure FillBuf(Buf: PStereoBuf; Channel, Freq: Integer; var Angle: Integer);
    procedure Start;
    procedure Stop;
  public
    { Public 錾 }
  end;

var
  DualWaveGen: TDualWaveGen;

implementation

{$R *.DFM}

procedure TDualWaveGen.FormCreate(Sender: TObject);
var
  Omega: double;
  I: Integer;
begin
  LVolume := 0.8;
  RVolume := 0.8;
  SamplingRate := 44100;   // CD quality
  SampleBits := 16;
  GetMem(FWaveBuf1, 4 * MaxBufSize);
  GetMem(FWaveBuf2, 4 * MaxBufSize);
  FreeBuf := FWaveBuf1;
  FillChar(FreeBuf^, 4 * MaxBufSize, 0);
  SetLength(LeftPoints, PaintBox1.Width);
  SetLength(RightPoints, PaintBox2.Width);
  SetLength(BeatPoints, PaintBox3.Width);
  SinTabCount := 14400;
  SetLength(SinTab, SinTabCount);
  Omega := 2 * Pi / SinTabCount;
  for I := 0 to SinTabCount -1 do
    SinTab[I] := Sin(Omega * I);
  GetMem(FWaveHdr1, SizeOf(WAVEHDR));
  GetMem(FWaveHdr2, SizeOf(WAVEHDR));
  MinFreq := 256;
  LeftFreq := MinFreq;
  RightFreq := MinFreq;

  {Wave Audio J}
  with FWaveFmt do
  begin
    wFormatTag := wave_Format_PCM;         // PCM data
    nChannels := 2;                        // 2 channels
    nSamplesPerSec := SamplingRate;        // set the 44.1KHz rate
    nAvgBytesPerSec := 4 * SamplingRate;   // four bytes per sample
    nBlockAlign := 4;                      // for 2-channel 16-bit audio
    wBitsPerSample := 16;                  // 16-bit audio
    cbSize := 0;
  end;
  ToneChange;
end;

procedure TDualWaveGen.FormDestroy(Sender: TObject);
begin
  FreeMem(FWaveHdr2);
  FreeMem(FWaveHdr1);
  SinTab := nil;
  FreeMem(FWaveBuf2);
  FreeMem(FWaveBuf1);
  BeatPoints := nil;
  RightPoints := nil;
  LeftPoints := nil;
end;

procedure TDualWaveGen.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  stop;
  shutoff := true;
end;

procedure TDualWaveGen.ToneChange;
begin
  EraseWave;
  SynthWave;
  DrawWave;
  SoundChanged:=true;
end;

procedure TDualWaveGen.SynthWave;
var
  I, J, H, W, BufPos: Integer;
  Rate: Double;
begin
  if not Playing then
  begin
    FillBuf(FWaveBuf1, 0, LeftFreq, LAngle);
    FillBuf(FWaveBuf1, 1, RightFreq, RAngle);
  end;
  BufPos := MulDiv(HrznScrollBar.Position, MaxBufSize, HrznScrollBar.Max);
  H := PaintBox1.Height div 2;
  W := PaintBox1.Width;
  Rate := H / 32767;
  if BufPos - W >= MaxBufSize then
    BufPos := MaxBufSize - W;
  for I := 0 to W - 1 do
  begin
    LeftPoints[I].X := I;
    J := 4 * I + BufPos;
    LeftPoints[I].Y := H - Trunc(Rate * FWaveBuf1[J]);
  end;
  H := PaintBox2.Height div 2;
  W := PaintBox2.Width;
  Rate := H / 32767;
  for I := 0 to W - 1 do
  begin
    RightPoints[I].X := I;
    J := 4 * I + BufPos;
    RightPoints[I].Y := H + Trunc(-Rate * FWaveBuf1[J+1]);
  end;
  W := PaintBox3.Width;
  H := PaintBox3.Height div 2;
  Rate := H / (2 * 32767);
  for I := 0 to W - 1 do
  begin
    BeatPoints[I].X := I;
    J := 4 * I + BufPos;
    BeatPoints[I].Y := H - Trunc(Rate * (FWaveBuf1[J] + FWaveBuf1[J + 1]));
  end;
end;


procedure TDualWaveGen.EraseWave;
begin
 // 
  with PaintBox1.Canvas do begin
    Pen.Color:=clTeal;
    PolyLine(LeftPoints);
  end;
  with PaintBox2.Canvas do begin
    Pen.Color:=clTeal;
    PolyLine(RightPoints);
  end;
  with PaintBox3.Canvas do begin
    Pen.Color:=clTeal;
    PolyLine(BeatPoints);
  end;
end;

procedure TDualWaveGen.DrawWave;
begin
 // `悷
  with PaintBox1.Canvas do begin
    Pen.Color:=clWhite;
    PolyLine(LeftPoints);
  end;
  with PaintBox2.Canvas do begin
    Pen.Color:=clWhite;
    PolyLine(RightPoints);
  end;
  with PaintBox3.Canvas do begin
    Pen.Color:=clWhite;
    PolyLine(BeatPoints);
  end;
end;

procedure TDualWaveGen.PlayBtnClick(Sender: TObject);
begin
  if Playing then
    Stop
  else
    Start;
end;

procedure TDualWaveGen.MMWomOpen(var Msg: TMessage);
begin
  with FWaveHdr1^ do begin
    lpData:= PAnsiChar(FWaveBuf1);
    dwBufferLength:= 4 * MaxBufSize;
    dwFlags:= 0;
    dwBytesRecorded := 0;
    dwUser := 0;
    dwLoops:= 0;
    lpNext := nil;
    reserved := 0;
  end;
  FWaveHdr2^ := FWaveHdr1^;
  FWaveHdr2^.lpData := PAnsiChar(FWaveBuf2);
  // FWaveHdȑA݁A㏈s
  WaveOutPrepareHeader(FWaveOut, FWaveHdr1, sizeof(WAVEHDR));
  Playing := True;
  waveOutWrite(FWaveOut, FWaveHdr1, SizeOf(WAVEHDR));
  WaveOutPrepareHeader(FWaveOut, FWaveHdr2, sizeof(WAVEHDR));
  if LeftFreq = RightFreq then
    RAngle := LAngle;
  FillBuf(FWaveBuf2, 0, LeftFreq, LAngle);
  FillBuf(FWaveBuf2, 1, RightFreq, RAngle);
  waveOutWrite(FWaveOut, FWaveHdr2, SizeOf(WAVEHDR));
end;

procedure TDualWaveGen.MMWomDone(var Msg: TMessage);
var
  Hdr: PWaveHdr;
begin
  if ShutOff then
  begin
    waveOutReset(FWaveOut);
    waveOutClose(FWaveOut);
  end
  else
  begin
    Hdr := PWaveHdr(Msg.LParam);
    FreeBuf := PStereoBuf(Hdr^.lpData);
    FillBuf(FreeBuf, 0, LeftFreq, LAngle);
    FillBuf(FreeBuf, 1, RightFreq, RAngle);
    waveOutWrite(FWaveOut, Hdr, SizeOf(WAVEHDR));
  end;
end;

procedure TDualWaveGen.MMWomClose(var Msg: TMessage);
begin
  WaveOutUnPrepareHeader(FWaveOut, FWaveHdr1, sizeof(WAVEHDR));
  WaveOutUnPrepareHeader(FWaveOut, FWaveHdr2, sizeof(WAVEHDR));
  FWaveOut := 0;
  Closed := True;
end;

procedure TDualWaveGen.Start;
var
  Status: MMRESULT;
begin
  if Playing then Stop;
  PlayBtn.Caption := 'Stop';
  Status := WaveOutOpen(@FWaveOut, 0, @FWaveFmt, Handle, 0, CALLBACK_WINDOW);
  if Status <> MMSYSERR_NOERROR then
  begin
    MessageDlg('EG[utH[I[fBIJ܂', mtWarning, [mbOk], 0);
    Exit;
  end;
  ShutOff := False;
  Playing := True;
end;

procedure TDualWaveGen.Stop;
begin
  if Playing then
  begin
    PlayBtn.Caption := 'Start';
    WaveOutReset(FWaveOut);
    Playing := False;
    ShutOff := True;
    Closed := False;
{
    repeat
      Application.ProcessMessages;
    until Closed;
}
  end;
end;

procedure TDualWaveGen.FillBuf(Buf: PStereoBuf; Channel, Freq: Integer; var Angle: Integer);
var
  Index, Step: Integer;
  Volume: double;
begin
  {16 bits, Stereo Wave Data 쐬āA炷}
  Step := Round(SinTabCount * Freq / SamplingRate);
  Index := 0;
  Volume := LVolume;
  if Channel > 0 then
  begin
    Volume := RVolume;
    Inc(Index);
  end;
  while Index < 2 * MaxBufSize do
  begin
    Buf^[Index] := (SmallInt(Trunc(32767 * Volume * SinTab[Angle])));
    Inc(Index, 2);
    Inc(Angle, Step);
    Angle := Angle mod SinTabCount;
  end;
end;


procedure TDualWaveGen.PaintBox1Paint(Sender: TObject);
begin
 // `悷
  DrawWave;
end;


procedure TDualWaveGen.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

procedure TDualWaveGen.LFreqBarChange(Sender: TObject);
begin
  with LFreqBar do
  begin
    LeftFreq := Max - Position + MinFreq;
    LFreqLabel.Caption:=IntToStr(LeftFreq);
  end;
  ToneChange;
end;

procedure TDualWaveGen.RFreqBarChange(Sender: TObject);
begin
  with RFreqBar do
  begin
    RightFreq := Max - Position + MinFreq;
    RFreqLabel.Caption:=IntToStr(RightFreq);
  end;
  ToneChange;
end;

procedure TDualWaveGen.HrznScrollBarChange(Sender: TObject);
begin
  ToneChange;
end;

procedure TDualWaveGen.LeftVolBarChange(Sender: TObject);
begin
  LVolume := LeftVolBar.Position / LeftVolBar.Max;
  ToneChange;
end;

procedure TDualWaveGen.RightVolBarChange(Sender: TObject);
begin
  RVolume := RightVolBar.Position / RightVolBar.Max;
  ToneChange;
end;

end.
